/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.jz.dmp.cmdexectool.scheduler.common.utils;

import java.sql.PreparedStatement;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.jz.dmp.cmdexectool.common.constant.CommConstant;
import com.jz.dmp.cmdexectool.scheduler.common.Constants;
import com.jz.dmp.cmdexectool.scheduler.common.enums.CommandType;
import com.jz.dmp.cmdexectool.scheduler.common.enums.DataType;
import com.jz.dmp.cmdexectool.scheduler.common.enums.MyDbType;
import com.jz.dmp.cmdexectool.scheduler.common.process.Property;
import com.jz.dmp.cmdexectool.scheduler.common.utils.placeholder.BusinessTimeUtils;
import com.jz.dmp.cmdexectool.scheduler.common.utils.placeholder.PlaceholderUtils;
import com.jz.dmp.cmdexectool.scheduler.common.utils.placeholder.TimePlaceholderUtils;

import freemarker.core.ReturnInstruction.Return;

/**
 * parameter parse utils
 */
public class ParameterUtils {

    private static final Logger logger = LoggerFactory.getLogger(ParameterUtils.class);

    private ParameterUtils() {
        throw new UnsupportedOperationException("Construct ParameterUtils");
    }

    /**
     * convert parameters place holders
     *
     * @param parameterString parameter
     * @param parameterMap parameter map
     * @return convert parameters place holders
     */
    public static String convertParameterPlaceholders(String parameterString, Map<String, String> parameterMap) {
        if (StringUtils.isEmpty(parameterString) || parameterMap == null) {
            return parameterString;
        }

        //Get current time, schedule execute time
        String cronTimeStr = parameterMap.get(Constants.PARAMETER_DATETIME);

        Date cronTime = null;

        if (StringUtils.isNotEmpty(cronTimeStr)) {
            cronTime = DateUtils.parse(cronTimeStr, Constants.PARAMETER_FORMAT_TIME);
        } else {
            cronTime = new Date();
        }

        // replace variable ${} form,refers to the replacement of system variables and custom variables
        parameterString = PlaceholderUtils.replacePlaceholders(parameterString, parameterMap, true);

        // replace time $[...] form, eg. $[yyyyMMdd]
        if (cronTime != null) {
            parameterString = TimePlaceholderUtils.replacePlaceholders(parameterString, cronTime, true);
        }

        return parameterString;
    }

    /**
     * new
     * convert parameters place holders
     *
     * @param parameterString parameter
     * @param parameterMap parameter map
     * @return convert parameters place holders
     */
    public static String convertParameterPlaceholders2(String parameterString, Map<String, String> parameterMap) {
        if (StringUtils.isEmpty(parameterString)) {
            return parameterString;
        }
        //Get current time, schedule execute time
        String cronTimeStr = parameterMap.get(Constants.PARAMETER_SHECDULE_TIME);
        Date cronTime = null;

        if (StringUtils.isNotEmpty(cronTimeStr)) {
            cronTime = DateUtils.parse(cronTimeStr, Constants.PARAMETER_FORMAT_TIME);

        } else {
            cronTime = new Date();
        }

        // replace variable ${} form,refers to the replacement of system variables and custom variables
        parameterString = PlaceholderUtils.replacePlaceholders(parameterString, parameterMap, true);

        // replace time $[...] form, eg. $[yyyyMMdd]
        if (cronTime != null) {
            parameterString = TimePlaceholderUtils.replacePlaceholders(parameterString, cronTime, true);

        }
        return parameterString;
    }

    /**
     * set in parameter
     *
     * @param index index
     * @param stmt preparedstatement
     * @param dataType data type
     * @param value value
     * @throws Exception errors
     */
    public static void setInParameter(int index, PreparedStatement stmt, DataType dataType, String value) throws Exception {
        if (dataType.equals(DataType.VARCHAR)) {
            stmt.setString(index, value);
        } else if (dataType.equals(DataType.INTEGER)) {
            stmt.setInt(index, Integer.parseInt(value));
        } else if (dataType.equals(DataType.LONG)) {
            stmt.setLong(index, Long.parseLong(value));
        } else if (dataType.equals(DataType.FLOAT)) {
            stmt.setFloat(index, Float.parseFloat(value));
        } else if (dataType.equals(DataType.DOUBLE)) {
            stmt.setDouble(index, Double.parseDouble(value));
        } else if (dataType.equals(DataType.DATE)) {
            stmt.setDate(index, java.sql.Date.valueOf(value));
        } else if (dataType.equals(DataType.TIME)) {
            stmt.setString(index, value);
        } else if (dataType.equals(DataType.TIMESTAMP)) {
            stmt.setTimestamp(index, java.sql.Timestamp.valueOf(value));
        } else if (dataType.equals(DataType.BOOLEAN)) {
            stmt.setBoolean(index, Boolean.parseBoolean(value));
        }
    }

    /**
     * curing user define parameters
     *
     * @param globalParamMap global param map
     * @param globalParamList global param list
     * @param commandType command type
     * @param scheduleTime schedule time
     * @return curing user define parameters
     */
    public static String curingGlobalParams(Map<String, String> globalParamMap, List<Property> globalParamList,
                                            CommandType commandType, Date scheduleTime) {

        if (globalParamList == null || globalParamList.isEmpty()) {
            return null;
        }

        Map<String, String> globalMap = new HashMap<>();
        if (globalParamMap != null) {
            globalMap.putAll(globalParamMap);
        }
        Map<String, String> allParamMap = new HashMap<>();
        //If it is a complement, a complement time needs to be passed in, according to the task type
        Map<String, String> timeParams = BusinessTimeUtils
                .getBusinessTime(commandType, scheduleTime);

        if (timeParams != null) {
            allParamMap.putAll(timeParams);
        }

        allParamMap.putAll(globalMap);

        Set<Map.Entry<String, String>> entries = allParamMap.entrySet();

        Map<String, String> resolveMap = new HashMap<>();
        for (Map.Entry<String, String> entry : entries) {
            String val = entry.getValue();
            if (val.startsWith("$")) {
                String str = ParameterUtils.convertParameterPlaceholders(val, allParamMap);
                resolveMap.put(entry.getKey(), str);
            }
        }
        globalMap.putAll(resolveMap);

        for (Property property : globalParamList) {
            String val = globalMap.get(property.getProp());
            if (val != null) {
                property.setValue(val);
            }
        }
        return JSONUtils.toJsonString(globalParamList);
    }

    /**
     * handle escapes
     *
     * @param inputString input string
     * @return string filter escapes
     */
    public static String handleEscapes(String inputString) {

        if (StringUtils.isNotEmpty(inputString)) {
            return inputString.replace("%", "////%").replaceAll("[\n|\r\t]", "_");
        }
        return inputString;
    }

    /**
     * $[yyyyMMdd] replace schedule time
     */
    public static String replaceScheduleTime(String text, Date scheduleTime) {
        Map<String, Property> paramsMap = new HashMap<>();
        //if getScheduleTime null ,is current date
        if (null == scheduleTime) {
            scheduleTime = new Date();
        }

        String dateTime = com.jz.dmp.cmdexectool.scheduler.common.utils.DateUtils.format(scheduleTime, Constants.PARAMETER_FORMAT_TIME);
        Property p = new Property();
        p.setValue(dateTime);
        p.setProp(Constants.PARAMETER_SHECDULE_TIME);
        paramsMap.put(Constants.PARAMETER_SHECDULE_TIME, p);
        text = ParameterUtils.convertParameterPlaceholders2(text, convert(paramsMap));

        return text;
    }

    /**
     * format convert
     *
     * @param paramsMap params map
     * @return Map of converted
     * see org.apache.dolphinscheduler.server.utils.ParamUtils.convert
     */
    public static Map<String, String> convert(Map<String, Property> paramsMap) {
        Map<String, String> map = new HashMap<>();
        Iterator<Map.Entry<String, Property>> iter = paramsMap.entrySet().iterator();
        while (iter.hasNext()) {
            Map.Entry<String, Property> en = iter.next();
            map.put(en.getKey(), en.getValue().getValue());
        }
        return map;
    }
    
    /**  
    * @Title: columnMappingHandler  
    * @Description: TODO(开发任务SQL：字段映射处理)  
    * @param @param jsonStr
    * @param @return    参数  
    * @return List<Map<String,String>>    返回类型  
    * @throws  
    */ 
    public static String columnMappingHandler(String jsonStr) {
    	JSONObject jsonObject = JSONObject.parseObject(jsonStr);
    	JSONArray sourceArray = jsonObject.getJSONArray("sourceFields");
    	Map<String, JSONObject> sourceMap = new HashMap<String, JSONObject>();
    	for (int index=0; index<sourceArray.size(); index++) {
			JSONObject sourceObj = sourceArray.getJSONObject(index);
			sourceMap.put(sourceObj.getString("customSoruceFieldId"), sourceObj);
		}
    	JSONArray targetArray = jsonObject.getJSONArray("targetFields");
    	Map<String, JSONObject> targetMap = new HashMap<String, JSONObject>();
    	for (int index=0; index<targetArray.size(); index++) {
			JSONObject targetObj = targetArray.getJSONObject(index);
			targetMap.put(targetObj.getString("customTargetFieldId"), targetObj);
		}
    	
    	StringBuilder sb = new StringBuilder("SELECT ");
    	
    	JSONArray mappingArray = jsonObject.getJSONArray("columnMapping");
    	Integer size = mappingArray.size();
    	for (int index=0; index<size; index++) {
    		JSONObject mappingObj = mappingArray.getJSONObject(index);
    		JSONObject sourceObj = sourceMap.get(mappingObj.getString("customSoruceFieldId"));
    		JSONObject targetObj = targetMap.get(mappingObj.getString("customTargetFieldId"));
    		
    		//String customSoruceFiledId = sourceObj.getString("customSoruceFiledId");
    		String sourceFieldName = sourceObj.getString("sourceFieldName");
    		//String sourceFieldType = sourceObj.getString("sourceFieldType");
    		//String customTargetFieldId = targetObj.getString("customTargetFieldId");
    		String targetFieldName = targetObj.getString("targetFieldName");
    		//String targetFieldType = targetObj.getString("targetFieldType");
    		
    		sb.append(sourceFieldName +" AS "+targetFieldName);
    		if (index<size-1) {
				sb.append(", ");
			}else {
				sb.append(" ");
			}
		}
    	sb.append("FROM t");
    	
    	return sb.toString();
    }
    
    /**  
     * @Title: columnMappingHandler  
     * @Description: TODO(开发任务SQL：生成主键冲突语句)  
     * @param @param jsonStr
     * @param @return    参数  
     * @return List<Map<String,String>>    返回类型  
     * @throws  
     */ 
     public static String columnMappingHandlerConflict(String jsonStr, String tableName, MyDbType myDbType) {
     	JSONObject jsonObject = JSONObject.parseObject(jsonStr);
     	JSONArray sourceArray = jsonObject.getJSONArray("sourceFields");
     	Map<String, JSONObject> sourceMap = new HashMap<String, JSONObject>();
     	for (int index=0; index<sourceArray.size(); index++) {
 			JSONObject sourceObj = sourceArray.getJSONObject(index);
 			sourceMap.put(sourceObj.getString("customSoruceFieldId"), sourceObj);
 		}
     	JSONArray targetArray = jsonObject.getJSONArray("targetFields");
     	Map<String, JSONObject> targetMap = new HashMap<String, JSONObject>();
     	for (int index=0; index<targetArray.size(); index++) {
 			JSONObject targetObj = targetArray.getJSONObject(index);
 			targetMap.put(targetObj.getString("customTargetFieldId"), targetObj);
 		}
     	
     	StringBuilder valuesSqlSb = new StringBuilder();
     	StringBuilder primaryKeySqlSb = new StringBuilder();
     	StringBuilder updateSqlSb = new StringBuilder();
     	
     	JSONArray mappingArray = jsonObject.getJSONArray("columnMapping");
     	Integer size = mappingArray.size();
     	for (int index=0; index<size; index++) {
     		JSONObject mappingObj = mappingArray.getJSONObject(index);
     		JSONObject sourceObj = sourceMap.get(mappingObj.getString("customSoruceFieldId"));
     		JSONObject targetObj = targetMap.get(mappingObj.getString("customTargetFieldId"));
     		
     		//String customSoruceFiledId = sourceObj.getString("customSoruceFiledId");
     		String sourceFieldName = sourceObj.getString("sourceFieldName");
     		//String sourceFieldType = sourceObj.getString("sourceFieldType");
     		//String customTargetFieldId = targetObj.getString("customTargetFieldId");
     		String targetFieldName = targetObj.getString("targetFieldName");
     		//String targetFieldType = targetObj.getString("targetFieldType");
     		String isPrimaryKey = targetObj.getString("isPrimaryKey");
     		
     		//设置values语句
 			valuesSqlSb.append("?");
 			
     		if (index!=size-1) {
     			valuesSqlSb.append(",");
			}
     		
     		if ("0".equals(isPrimaryKey) && myDbType.MySQL==myDbType) {
     			
     			//设置update语句
     			updateSqlSb.append(targetFieldName);
     			updateSqlSb.append("=VALUES(");
     			updateSqlSb.append(targetFieldName);
     			updateSqlSb.append(")");
     			updateSqlSb.append(",");
     			
			} else if (myDbType.PostgreSQL==myDbType) {
				if ("1".equals(isPrimaryKey)) {
					primaryKeySqlSb.append("targetFieldName");
					primaryKeySqlSb.append(",");
				}else {
					//设置update语句
	     			updateSqlSb.append(targetFieldName);
	     			updateSqlSb.append("=EXCLUDED.");
	     			updateSqlSb.append(targetFieldName);
	     			updateSqlSb.append(",");
				}
			}
 		}
     	
     	String primaryKeySqlStr = "";
     	if (org.apache.commons.lang3.StringUtils.isNotEmpty(primaryKeySqlSb)) {
     		primaryKeySqlStr = primaryKeySqlSb.substring(0, primaryKeySqlSb.length()-1);
		}
     	String updateSqlStr = "";
     	if (org.apache.commons.lang3.StringUtils.isNotEmpty(updateSqlSb)) {
     		updateSqlStr = updateSqlSb.substring(0, updateSqlSb.length()-1);
		}
     	
     	StringBuilder sb = new StringBuilder();
     	
     	if (myDbType.MySQL==myDbType) {
     		sb.append("INSERT INTO ");
         	sb.append(tableName+" ");
         	sb.append("VALUES(");
         	sb.append(valuesSqlSb);
         	sb.append(") ");
         	
     		sb.append("ON DUPLICATE KEY UPDATE ");
     		sb.append(updateSqlStr);
		} else if (myDbType.PostgreSQL==myDbType) {
			sb.append("INSERT INTO ");
	     	sb.append(tableName+" ");
	     	sb.append("VALUES(");
	     	sb.append(valuesSqlSb);
	     	sb.append(") ");
	     	
			sb.append("ON CONFLICT(");
			sb.append(primaryKeySqlStr);
			sb.append(") do update ");
			sb.append(updateSqlStr);
		}
     	
     	
     	return sb.toString();
     }

}
